@@ -1,29 +1,22 @@
+import asyncio
 from typing import AsyncGenerator, Optional, Any
 
 import aiohttp
 
+from simple_avk import exceptions
+
 GROUPS_LONGPOLL_METHOD = "groups.getLongPollServer"
 USERS_LONGPOLL_METHOD = "messages.getLongPollServer"
-LONGPOLL_ERROR_DESCRIPTIONS = {
-    1: (
-        "Events history is outdated or partially lost, app "
-        "can get further events using new ts value from server response."),
-    2: (
-        "Key is outdated, you need to get new key "
-        "using method {}."),
-    3: (
-        "Information about user is lost, you need to request "
-        "new key and ts using method {}."),
-    4: (
-        "Invalid version number passed in 'version' parameter.")
-}
-LONGPOLL_ERROR_MSG = "Error in longpoll with code {}: {}"
-VK_METHOD_ERROR_MSG = "Error in method {} with code {}: {}"
+
+FOURTH_LONGPOLL_ERROR_DESCRIPTION = (
+    "Invalid version number passed in 'version' parameter."
+)
+UNKNOWN_LONGPOLL_ERROR_MSG = "Unknown longpoll error!"
+
 VK_METHOD_LINK = "https://api.VK.com/method/{}"
 
 
 class SimpleAVK:
-
     """
     Main class of the simple_avk library.
 
@@ -70,6 +63,13 @@ class SimpleAVK:
         self.longpoll_server_link = ""
         self.longpoll_params = {}
 
+    async def _get_new_longpoll_info(self) -> dict:
+        return await self.call_method(
+            self.longpoll_method,
+            # Line too long cuz pep8 symbol limit sucks
+            params={} if self.group_id is None else {"group_id": self.group_id}
+        )
+
     async def _prepare_longpoll(self) -> None:
         """
         Getting longpoll server to receive events.
@@ -77,17 +77,11 @@ class SimpleAVK:
         Raises:
            VKError: any error from VK response
         """
-        if self.group_id:
-            self.longpoll_method = GROUPS_LONGPOLL_METHOD
-            longpoll_params = {"group_id": self.group_id}
-        else:
-            # If group id isn't specified - it's user longpoll
-            self.longpoll_method = USERS_LONGPOLL_METHOD
-            longpoll_params = {}
-        resp = await self.call_method(
-            self.longpoll_method,
-            params=longpoll_params
+        # If group id isn't specified - it's user longpoll
+        self.longpoll_method = (
+            GROUPS_LONGPOLL_METHOD if self.group_id else USERS_LONGPOLL_METHOD
         )
+        resp = await self._get_new_longpoll_info()
         vk_last_event_id = resp["ts"]
         vk_secret_key = resp["key"]
         vk_longpoll_server = resp["server"]
@@ -110,6 +104,42 @@ class SimpleAVK:
                 }
             )
 
+    async def _real_get_new_events(self) -> list:
+        updates = None
+        while updates is None:
+            try:
+                resp = await self.aiohttp_session.get(
+                    self.longpoll_server_link,
+                    params=self.longpoll_params
+                )
+            except (aiohttp.ServerDisconnectedError, asyncio.TimeoutError):
+                continue
+            resp_json = await resp.json()
+            del resp
+            if "failed" in resp_json:
+                error_code = resp_json["failed"]
+                if error_code == 1:  # Events are partially lost or outdated
+                    # This error gives a new ts
+                    self.longpoll_params["ts"] = resp_json["ts"]
+                elif error_code in (2, 3):  # 2 - key is outdated
+                    new_server_info = await self._get_new_longpoll_info()
+                    self.longpoll_params["key"] = new_server_info["key"]
+                    if error_code == 3:  # User info lost
+                        # This error gives a new ts
+                        self.longpoll_params["ts"] = new_server_info["ts"]
+                elif error_code == 4:
+                    raise exceptions.LongpollError(
+                        error_code, FOURTH_LONGPOLL_ERROR_DESCRIPTION
+                    )
+                else:
+                    raise exceptions.LongpollError(
+                        error_code, UNKNOWN_LONGPOLL_ERROR_MSG
+                    )
+            else:
+                self.longpoll_params["ts"] = resp_json["ts"]
+                updates = resp_json["updates"]
+        return updates
+
     async def get_new_events(self) -> list:
         """
         Get new events (also called updates) from longpoll.
@@ -120,24 +150,11 @@ class SimpleAVK:
         Raises:
             VKError: any error from VK longpoll
         """
-        if not self.longpoll_server_link:
-            await self._prepare_longpoll()
-        resp = await self.aiohttp_session.get(
-            self.longpoll_server_link,
-            params=self.longpoll_params
-        )
-        resp_json = await resp.json()
-        if "failed" not in resp_json:
-            if "ts" in resp_json:
-                self.longpoll_params["ts"] = resp_json["ts"]
-            return resp_json["updates"]
-        error_code = resp_json["failed"]
-        raise LongpollError(
-            error_code,
-            LONGPOLL_ERROR_DESCRIPTIONS[error_code].format(
-                self.longpoll_method
-            )
-        )
+        await self._prepare_longpoll()
+        # noinspection PyAttributeOutsideInit
+        # because I'm changing a function to make something like state pattern
+        self.get_new_events = self._real_get_new_events
+        return await self._real_get_new_events()
 
     async def listen(self) -> AsyncGenerator[Any, None]:
         """
@@ -183,44 +200,6 @@ class SimpleAVK:
         if "error" not in resp_json:
             return resp_json["response"]
         error = resp_json["error"]
-        raise MethodError(
+        raise exceptions.MethodError(
             method_name, error["error_code"], error["error_msg"]
         )
-
-
-class MethodError(Exception):
-
-    """
-    Exception.
-    Raised when an error is received from the VK method response.
-    """
-
-    def __init__(self, method_name: str, error_code: int, message: str) -> None:
-        self.method_name = method_name
-        self.error_code = error_code
-        self.message = message
-
-    def __str__(self) -> str:
-        return VK_METHOD_ERROR_MSG.format(
-            self.method_name,
-            self.error_code,
-            self.message
-        )
-
-
-class LongpollError(Exception):
-
-    """
-    Exception.
-    Raised when an error is received from the VK longpoll.
-    """
-
-    def __init__(self, error_code: int, message: str) -> None:
-        self.error_code = error_code
-        self.message = message
-
-    def __str__(self) -> str:
-        return LONGPOLL_ERROR_MSG.format(
-            self.error_code,
-            self.message
-        )
